unit Scanner;
{$R-,B-,I-}
interface
uses
   Objects,
   Dirs;

const
   TestCluster1 =  9;  { 512 }
   TestCluster2 =  14; { 16384 }

type
   TConfigFile =
   object(TObject)
      Exists       :Boolean;

      FileName     :PString;
      Data         :PCollection;

      Line         :PString;
      LineNum      :Integer;

      constructor Init(const AFileName :String);

      procedure Read;

      function  IsLine(const CmdMask   :String)  :Boolean;
   end;

   TScanDisk =
   object(TProcessFiles)
      InvExe       :TCollection;

      TempFiles    :TCollection;
      TempSize     :LongInt;

      ClusterSize  :Word;
      ClustersCount:Word;
      UsedSize,
      UsedSizeC    :LongInt;
      FilesCount   :LongInt;

      UsedSizeCT   :Array [TestCluster1..TestCluster2] of LongInt;

      Config_Sys,
      AutoExec_Bat :^TConfigFile;

      Mft_Exe,
      StDbl_Sys    :PString;

      constructor Init;

      procedure Run;
      procedure ProcessFile; virtual;

      procedure ShowReport;
   end;

implementation

{ TConfigFile }

constructor TConfigFile.Init;
begin
   Inherited Init;

   FileName:=NewStr(AFileName);
end;

procedure TConfigFile.Read;
var
   f               :Text;
   Str             :String;

procedure NormalizeStr(var Str         :String);
var
   p               :Integer;
begin
   asm
      les  di,Str
      xor  cx,cx
      mov  cl,es:[di]
      jcxz @@3
@@1:
      inc  di
      mov  al,es:[di]
      cmp  al,9
      jne  @@4
      mov  al,' '
@@4:
      cmp  al,'a'
      jb   @@2
      cmp  al,'z'
      ja   @@2
      sub  al,'a'-'A'
      mov  es:[di],al
@@2:
      loop @@1
@@3:
   end;

   while Str[1]=' ' do Delete(Str, 1, 1);
   while (Str<>'') and (Str[Length(Str)]=' ') do Dec(Str[0]);
   repeat
      p:=Pos('  ', Str);
      if p<>0 then Delete(Str, p, 1);
   until p=0;
end;

begin
   if Data<>nil then Exit;
   New(Data, Init(0, 16));
   Assign(f, FileName^);
   Reset(f);
   while not EOF(f) do
   begin
      ReadLn(f, Str);
      NormalizeStr(Str);
      if (Copy(Str, 1, 4)<>'REM ') and (Str[1]<>':') and (Str[1]<>';') and (Str<>'') then
         Data^.Insert(NewStr(Str));
   end;
   Close(f);
   Exists:=(IOResult=0);
end;

function  TConfigFile.IsLine;
var
   i               :Integer;
   Eq              :Boolean;

function  Right(const S                :String;
                n                      :Integer) :String;
var
   l               :Integer;
begin
   l:=Length(S);
   if n>l then n:=l;
   Right:=Copy(S, l-n+1, n);
end;

begin
   if Data<>nil then
   for i:=1 to Data^.Count do
   begin
      LineNum:=i;
      Line:=Data^.At(i-1);

      if CmdMask[Length(CmdMask)]='*' then if CmdMask[1]='*'
      then Eq:=(Pos(Copy(CmdMask, 2, Length(CmdMask)-2), ' '+Line^+' ')<>0)
      else Eq:=(Copy(CmdMask, 1, Length(CmdMask)-1)=Copy(Line^, 1, Length(CmdMask)-1))
      else if CmdMask[1]='*'
      then Eq:=(Right(CmdMask, Length(CmdMask)-1)=Right(Line^, Length(CmdMask)-1))
      else Eq:=(CmdMask=Line^);

      if Eq then
      begin
         IsLine:=True; Exit;
      end;
   end;
   IsLine:=False;
end;

{ TScanDisk }

constructor TScanDisk.Init;
begin
   Inherited Init;

   asm
      mov  ah,1Bh
      push ds
      int  21h
      pop  ds
      les  di,Self
      mov  es:[di].TScanDisk.ClustersCount,dx
      xor  ah,ah
      mul  cx
      mov  es:[di].TScanDisk.ClusterSize,ax
   end;

   InvExe.Init(64, 64);
   TempFiles.Init(64, 256);

   New(Config_Sys, Init('\CONFIG.SYS'));
   New(AutoExec_Bat, Init('\AUTOEXEC.BAT'));
end;

procedure TScanDisk.Run;
begin
   WriteLn('Scanning disk ...'#13#10);
   Inherited Run('\');
end;

procedure TScanDisk.ProcessFile;
type
   TExeHeader =
   record
      MZ           :Array [0..1] of Char;
      PartPage,
      PageCnt      :Word;
   end;

var
   Ext             :String[3];
   p               :Integer;

   RealSize        :LongInt;

   S               :TDosStream;

   ExeHeader       :TExeHeader;
   SysHeader       :LongInt absolute ExeHeader;
   CodeSize        :LongInt;
   ExeId,
   InvHeader       :Boolean;

   TestClusterI    :TestCluster1..TestCluster2;

const
   LTimer          :Byte =   0;
var
   Timer           :Byte absolute $0040:$006C;

function  CalcRealSize(ClusterSize     :Word)    :LongInt;
begin
   if Info.Size mod ClusterSize = 0
   then CalcRealSize:=Info.Size
   else CalcRealSize:=Info.Size - (Info.Size mod ClusterSize) + ClusterSize;
end;

begin
   RealSize:=CalcRealSize(ClusterSize);

   for TestClusterI:=TestCluster1 to TestCluster2 do
      Inc(UsedSizeCT[TestClusterI], CalcRealSize(1 shl TestClusterI));

   Inc(FilesCount);
   Inc(UsedSize, Info.Size);
   Inc(UsedSizeC, RealSize);

   p:=Pos('.', Info.Name);
   if p<>0 then Ext:=Copy(Info.Name, p+1, 255) else Ext:='';

   if Info.Name='ST-DBL.SYS' then StDbl_Sys:=NewStr(Dir+Info.Name) else
   if Info.Name='MFT.EXE' then Mft_Exe:=NewStr(Dir+Info.Name);

   if (Ext='EXE') or (Ext='DLL') or (Ext='COM') or
      ((Ext='SYS') and (Info.Name<>'CONFIG.SYS') and
      (Info.Name<>'MSDOS.SYS') and (Info.Name<>'IO.SYS')) then
   begin
      S.Init(Dir+Info.Name, stOpenRead);

      if S.GetSize>=SizeOf(ExeHeader)
      then S.Read(ExeHeader, SizeOf(ExeHeader))
      else ExeHeader.MZ:=#0#0;

      ExeId:=(ExeHeader.MZ='MZ') or (ExeHeader.MZ='ZM');
      if Ext='SYS' then
      begin
         InvHeader:=(SysHeader<>$FFFFFFFF);
         if ExeId or (InvHeader and (Byte(SysHeader)=$FF)) then InvHeader:=False;
      end
      else
      begin
         CodeSize:=LongMul(ExeHeader.PageCnt, 512);
         if ExeHeader.PartPage<>0 then Dec(CodeSize, 512 - ExeHeader.PartPage);

         InvHeader:=ExeId and (CodeSize>S.GetSize);
      end;

      S.Done;
      if S.Status<>stOk then Status:=rdError;

      if (((Ext='EXE') or (Ext='DLL')) and (not ExeId)) or
         ((Ext='COM') and ExeId and (not (Dir='\DOS\'))) or
         InvHeader
      then InvExe.Insert(NewStr(Dir+Info.Name));
   end else
   if (Ext='TMP') or (Ext='$$$') or (Info.Name='WIN386.SWP') then
   begin
      TempFiles.Insert(NewStr(Dir+Info.Name));
      Inc(TempSize, RealSize);
   end else
   if Dir='\' then
   if Info.Name='CONFIG.SYS' then Config_Sys^.Read else
   if Info.Name='AUTOEXEC.BAT' then AutoExec_Bat^.Read;

   if IOResult<>0 then Status:=rdError;

   if Timer<>LTimer then
   begin
      Write(#13, FilesCount:5, ' files');
      LTimer:=Timer;
   end;
end;

procedure TScanDisk.ShowReport;
var
   i               :Integer;

   AverageSize     :LongInt;

   TestClusterI    :TestCluster1..TestCluster2+1;

   TestCluster,
   BestCluster     :Word;

   CLostSize,
   LostSize,
   MinLostSize     :LongInt;

   Comments        :Word;

   QEMM,
   EMM386, HiMem,
   HMA             :Boolean;

   WriteCache      :Boolean;

begin
   WriteLn(
      #13#10#9#9#9#9' Report'+
      #13#10#9#9#9#9''#13#10);

   Comments:=0;
   WriteLn(#9'Configuration files :'#13#10);

   with Config_Sys^ do
   if Exists then
   begin
      QEMM:=IsLine('*QEMM386*');
      EMM386:=IsLine('*EMM386.EXE*');
      HiMem:=IsLine('*HIMEM.SYS*');

      HMA:=HiMem or QEMM or IsLine('*386MAX*') or IsLine('*LASTBYTE*');

      if not IsLine('STACKS=0*') then
      begin
         Inc(Comments);
         WriteLn('Stop allocating DOS stacks.');
         if IsLine('STACKS=*')
         then Write('Change the line ''', Line^, ''' to ''STACKS=0,0'' in')
         else Write('Add the line ''STACKS=0,0'' to');
         WriteLn(' your CONFIG.SYS.'#13#10);
      end;

      if not IsLine('FCBS=1,0*') then
      begin
         Inc(Comments);
         WriteLn('Add the line ''FCBS=1,0'' to your CONFIG.SYS.'#13#10);
      end;

      if not IsLine('LASTDRIVE=*') then
      begin
         Inc(Comments);
         WriteLn('Add the line LASTDRIVE=X: to your CONFIG.SYS, where X is the last drive'#13#10+
            'letter you may ever need.'#13#10);
      end;

      if HMA and (not IsLine('DOS=HIGH*')) then
      begin
         Inc(Comments);
         Write('Add the line ''DOS=HIGH');
         if not QEMM then Write(',UMB');
         WriteLn(''' to your CONFIG.SYS.'#13#10);
      end;

      if Test8086>=2 then
      begin
         if QEMM then
         begin
            if HiMem then
            begin
               Inc(Comments);
               WriteLn('Stop loading HIMEM.SYS. It is not needed when you use QEMM.'#13#10);
            end;
         end else
         begin
            Inc(Comments);
            Write('You have at least an 80386 CPU. Why don''t you install QEMM386.SYS'#13#10+
               'on your machine? ');
            if EMM386 then
            WriteLn('It''s better than EMM386.EXE that you are currently using.');
            WriteLn('Run the PC-Advisor again after you install QEMM.'#13#10);
         end;

         if IsLine('*DBLSPACE*') and (not IsLine('*ST-DBL*')) then
         begin
            Inc(Comments);
            WriteLn('You are using DoubleSpace. You can load it into the expanded memory'#13#10+
               'using the Quarterdeck''s ST-DBL.SYS driver.');
            if StDbl_Sys<>nil then
            WriteLn('This driver is found on your disk: ', StDbl_Sys^, '.') else
            WriteLn('There''s no ST-DBL.SYS driver on your disk. You should install all the files'#13#10+
               'of QEMM 7.0 or later version in order to use this driver.');
            WriteLn('It''ll be better if you use the QSETUP.EXE program to install the driver.');
            if not QEMM then WriteLn('But install QEMM first!');
            WriteLn;
         end;
      end;
   end;

   with AutoExec_Bat^ do
   if Exists then
   begin
      if IsLine('*SMARTDRV*') then
      begin
         WriteCache:=True;
         i:=Pos('SMARTDRV', Line^);
         while (i<=Length(Line^)) and (Line^[i]<>' ') do Inc(i);
         while i<Length(Line^) do
         begin
            if ((Line^[i]>='A') and (Line^[i]<='Z') and
               (Line^[i-1]=' ') and ((Line^[i+1]=' ') or (Line^[i+1]='-'))) or
               ((Line^[i]='/') and (Line^[i+1]='X'))
            then WriteCache:=False;
            Inc(i);
         end;
         if WriteCache then
         begin
            Inc(Comments);
            WriteLn('You''re loading SMARTDRV.EXE from your AUTOEXEC.BAT with enabled'#13#10+
               'write-caching. This may cause loss of data. I recommend you to turn');
            WriteLn('the write-caching off. To do it place your hard disk drive letter after the'#13#10+
               'SMARTDRV command. Run SMARTDRV /? for more information.'#13#10);
         end;
      end;

      if IsLine('*PU_1700*') and (Pos('/SM=', Line^)=0) then
      begin
         Inc(Comments);
         WriteLn('You''re loading PU_1700.COM from your AUTOEXEC.BAT. Why don''t you use the'#13#10+
            '/SM=1 or /SM=2 options? This may increase the speed of your floppy drives.'#13#10);
      end;
   end;

   if Mft_Exe<>nil then
   begin
      Inc(Comments);
      WriteLn('Quarterdeck MANIFEST is found on your disk: ', Mft_Exe^, '.'#13#10+
         'Run it, select ''Hints'' from the main menu and see what it can advise you.'#13#10);
   end;

   if Comments=0 then
   WriteLn('No advices.'#13#10);

   if InvExe.Count<>0 then
   begin
      WriteLn(#9'Invalid executables :'#13#10);
      for i:=0 to InvExe.Count-1 do WriteLn(PString(InvExe.At(i))^);
      if InvExe.Count=1
      then Write(#13#10'This file')
      else Write(#13#10'These files');
      Write(' or the programs that use ');
      if InvExe.Count=1
      then Write('it')
      else Write('them');
      Write(' may not work properly or may cause'#13#10'other programs '+
         'not to work properly. So if you have problems with running '#13#10+
         'some programs, try to reinstall ');
      if InvExe.Count=1
      then Write('this file')
      else Write('these files');
      WriteLn('.'#13#10);
   end;

   if TempFiles.Count<>0 then
   begin
      WriteLn(#9'Temporary files :'#13#10);
      for i:=0 to TempFiles.Count-1 do WriteLn(PString(TempFiles.At(i))^);
      if TempFiles.Count=1
      then Write(#13#10'This file is')
      else Write(#13#10'These files are');
      if TempSize=0
      then Write('n''t occupying any disk space.'#13#10'Nevertheless, y')
      else
      begin
         Write(' occupying ', TempSize shr 10, ' Kb of disk space, but ');
         if TempFiles.Count=1
         then Write('it is')
         else Write('they are');
         Write(' no longer needed.'#13#10'Y');
      end;
      Write('ou''d better delete ');
      if TempFiles.Count=1
      then Write('it')
      else Write('them');
      WriteLn('.'#13#10);
   end;

   if FilesCount=0 then AverageSize:=0 else AverageSize:=UsedSize div FilesCount;
   WriteLn(#9'Useful information :'#13#10#13#10+
      'Total files on disk :'#9#9, FilesCount, #13#10+
      'Total disk size :'#9#9, (LongInt(ClustersCount) * ClusterSize) shr 10, ' Kb'#13#10+
      'Used space (files) :'#9#9, UsedSize shr 10, ' Kb'#13#10+
      'Used space (in reality) :'#9, UsedSizeC shr 10, ' Kb'#13#10+
      'Average file size :'#9#9, AverageSize, ' bytes'#13#10+
      'Cluster size :'#9#9#9, ClusterSize, ' bytes'#13#10);

   CLostSize:=(UsedSizeC-UsedSize) + LongInt(ClustersCount) shl 1;

   WriteLn(
      'The following table may help you choose the cluster size:'#13#10#13#10+
      #9'Cluster size'#9#9'Lost space');

   TestClusterI:=TestCluster1;
   TestCluster:=1 shl TestCluster1; BestCluster:=1 shl TestCluster1;
   MinLostSize:=$7FFFFFFF;
   repeat
      LostSize:=
         (UsedSizeCT[TestClusterI]-UsedSize) +
         ((LongInt(ClustersCount) * ClusterSize div TestCluster) shl 1);

      Write(#9, TestCluster:9, #9#9, (LostSize shr 10):6, ' Kb');
      if (TestCluster<ClusterSize) and
         (LongInt(ClustersCount) * ClusterSize div TestCluster>$FFFF)
      then WriteLn(#9' (not supported by DOS)')
      else WriteLn;

      if LostSize<MinLostSize then
      begin
         MinLostSize:=LostSize; BestCluster:=TestCluster;
      end;
      TestCluster:=TestCluster shl 1; Inc(TestClusterI);
   until TestClusterI>TestCluster2;
   WriteLn;

   with Config_Sys^ do
   if IsLine('*DBLSPACE*') or IsLine('*ST-DBL*') or IsLine('*STACKER*') then
   WriteLn('WARNING: The results shown in the table above may not be very accurate:'#13#10+
      'it seems like your drive is compressed by DoubleSpace or Stacker.'#13#10);

   if (BestCluster=ClusterSize) or (CLostSize-MinLostSize<1024)
   then WriteLn('You''re currently using the best cluster size for your data.')
   else
   begin
      WriteLn('The best cluster size for your data is ', BestCluster, ' bytes.'#13#10+
         'Using this cluster size you''ll get ', (CLostSize-MinLostSize) shr 10,
         ' Kb more free disk space.');
      if BestCluster<4096 then
      WriteLn('You may want to use a little larger cluster size not to make your disk'#13#10+
         'too fragmented and too slow.');
   end;

   WriteLn(
      #13#10''+
      #13#10' PC-Advisor by Solar Designer');
end;

end.

